#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : AI.  @by PyCharm
# @File         : file
# @Time         : 2023/7/15 17:39
# @Author       : betterme
# @WeChat       : meutils
# @Software     : PyCharm
# @Description  :

from meutils.pipe import *
from langchain.text_splitter import *
from langchain.document_loaders import CSVLoader, UnstructuredFileLoader
from langchain.document_loaders.base import BaseLoader
from langchain.schema import Document
from chatllm.llmchain.document_loaders import TextLoader, Docx2txtLoader, PyMuPDFLoader


class FileLoader(BaseLoader):
    """
        loader = FilesLoader('data/古今医统大全.txt')
        docs = loader.load_and_split()
    """

    def __init__(
        self,
        file_path: Any,
        filename: Optional[str] = None,
        max_workers: int = 3
    ):
        self.file_path = file_path

        self.filename = filename or file_path
        if not isinstance(self.filename, str):
            if hasattr(self.filename, 'name'):
                self.filename = self.filename.name
            else:
                self.filename = str(self.filename)

        self._max_workers = max_workers

    # @staticmethod
    # def _load_one_file(file_args, text_splitter=None):
    #     file_stream, filename = file_args
    #     docs = FileLoader(file_stream, filename).load_and_split(text_splitter)
    #     print(docs)
    #     for idx, doc in enumerate(docs):
    #         doc.metadata['index'] = idx
    #         doc.metadata['source'] = doc.metadata.get('source') or filename
    #     return docs or []
    #
    # @classmethod
    # def load_and_split4multifile(cls, files, text_splitter=None):
    #     _load_one_file = partial(cls._load_one_file, text_splitter=text_splitter)
    #     docs = files | xmap(lambda file: (file.read(), file.name)) | xProcessPoolExecutor(_load_one_file, min(8, len(files))) | xchain_
    #     return docs

    def load_and_split(
        self,
        text_splitter: Optional[TextSplitter] = None,
    ) -> List[Document]:
        """Load Documents and split into chunks. Chunks are returned as Documents.

        Args:
            text_splitter: TextSplitter instance to use for splitting documents.
              Defaults to RecursiveCharacterTextSplitter.

        Returns:
            List of Documents.
        """
        if text_splitter is None:
            separators = ['\n\n', '\r', '\n', '\r\n', '。', '!', '！', '\\?', '？', '……', '…']
            _text_splitter = RecursiveCharacterTextSplitter(
                chunk_size=510,
                chunk_overlap=100,
                add_start_index=True,
                separators=separators
            )
        else:
            _text_splitter = text_splitter
        docs = self.load()
        return _text_splitter.split_documents(docs)

    def load(self) -> List[Document]:

        if self.filename.endswith((".txt",)):
            docs = TextLoader(self.file_path, autodetect_encoding=True).load()

        elif self.filename.endswith((".docx",)):
            docs = Docx2txtLoader(self.file_path).load()

        elif self.filename.endswith((".pdf",)):
            docs = PyMuPDFLoader(self.file_path).load()

        elif self.filename.endswith((".csv",)):
            docs = CSVLoader(self.file_path).load()

        else:
            if isinstance(self.file_path, List) or (
                isinstance(self.file_path, (str, os.PathLike))
                and len(self.file_path) < 256
                and Path(self.file_path).is_file()
            ):
                docs = UnstructuredFileLoader(
                    file_path=self.file_path,
                    strategy="fast",
                ).load()
            else:
                docs = UnstructuredFileLoader(
                    file_path=None,
                    file=self.file_path,
                    file_filename=self.filename,
                    strategy="fast",
                ).load()
                for doc in docs:
                    doc.metadata['source'] = doc.metadata.get('source') or self.filename

        # schema: file_type todo: 增加字段
        # 静态schema怎么设计存储，支持多文档：metadata存文件名字段（可以放多层级）
        # docs[0].metadata['total_length'] = len(docs[0].page_content)
        # docs[0].metadata['file_name'] = Path(docs[0].metadata['source']).name
        # docs[0].metadata['ext'] = {}  # 拓展字段

        return docs


if __name__ == '__main__':
    p = 'text.py'
    # p = '/Users/betterme/PycharmProjects/AI/ChatLLM/data/医/古今医统大全.txt'
    loader = FileLoader(p)
    docs = loader.load_and_split()
    print(docs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : AI.  @by PyCharm
# @File         : closeai
# @Time         : 2023/5/24 14:38
# @Author       : betterme
# @WeChat       : meutils
# @Software     : PyCharm
# @Description  : todo: 整合所有api
"""
openai
chatglm
ziya
bard
文心
讯飞
"""

from meutils.pipe import *
import openai

openai.api_base = 'https://api.openai-proxy.com/v1'
openai.api_key = "sk-xx"  # supply your API key however you choose


def create(prompt="1+1", stream=True, max_tokens=2048, model="text-davinci-003"):
    completion = openai.Completion.create(model=model, prompt=prompt, stream=stream, max_tokens=max_tokens)
    if not stream:
        yield from completion.choices[0].text
    for c in completion:
        yield c.choices[0].text

from langchain.retrievers import BM25Retriever, EnsembleRetriever

from langchain.retrievers.document_compressors import LLMChainExtractor

from langchain.storage import InMemoryStore

from langchain.agents.agent_toolkits import OpenAPIToolkit




#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : AI.  @by PyCharm
# @File         : prompt_maker
# @Time         : 2023/11/10 17:09
# @Author       : betterme
# @WeChat       : meutils
# @Software     : PyCharm
# @Description  :

from meutils.pipe import *

from langchain import hub
obj = hub.pull("hardkothari/prompt-maker")

from langchain.chains import LLMChain
from langchain.chat_models import ChatOpenAI

llm = LLMChain(llm=ChatOpenAI(), prompt=obj)
llm.run(lazy_prompt='拖把', task='小红书写文章, 用中文回答')




#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : AI.  @by PyCharm
# @File         : callbacks
# @Time         : 2023/7/12 17:37
# @Author       : betterme
# @WeChat       : meutils
# @Software     : PyCharm
# @Description  :
from threading import Thread

from langchain.chat_models import ChatOpenAI as _ChatOpenAI
from langchain.callbacks import AsyncIteratorCallbackHandler
from chatllm.llmchain.callbacks import StreamingGeneratorCallbackHandler

from meutils.pipe import *

# llm._get_llm_string()
class ChatOpenAI(_ChatOpenAI):

    def stream(self, text: str, *, stop: Optional[Sequence[str]] = None, **kwargs: Any) -> Generator:
        """Stream the answer to a query.

        NOTE: this is a beta feature. Will try to build or use
        better abstractions about response handling.

        """
        # if self.cache: return iter(self.predict(text, stop=stop, **kwargs)) # 在外面做缓存逻辑

        handler = StreamingGeneratorCallbackHandler()
        self.callbacks = [handler]
        self.streaming = True

        #  background_tasks.add_task(_predict, text, **kwargs)
        kwargs['stop'] = stop
        thread = Thread(target=self.predict, args=[text], kwargs=kwargs)
        thread.start()  # thread.is_alive() 瞬间完成从缓存里取

        return handler.get_response_gen()

    async def astream(self, text: str, *, stop: Optional[Sequence[str]] = None, **kwargs: Any) -> AsyncGenerator:
        handler = AsyncIteratorCallbackHandler()
        self.callbacks = [handler]
        self.streaming = True

        task = asyncio.create_task(self.apredict(text, stop=stop, **kwargs))

        async for token in handler.aiter():
            yield token



if __name__ == '__main__':
    llm = ChatOpenAI(streaming=True, temperature=0)
    for i in llm.stream('你好'):
        print(i, end='')

    async for token in llm.astream('你好'):
        print(token, end='')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : AI.  @by PyCharm
# @File         : visualglm_st.py
# @Time         : 2023/5/26 15:38
# @Author       : betterme
# @WeChat       : meutils
# @Software     : PyCharm
# @Description  :

from meutils.pipe import *
from transformers import AutoModel, AutoTokenizer
import streamlit as st
from PIL import Image
import numpy as np
import tempfile
from streamlit_chat import message as message_chat


def predict(input, image_path, chatbot, max_length, top_p, temperature, history):
    if image_path is None:
        return [(input, "图片为空！请重新上传图片并重试。")]
    for response, history in st.session_state.model.stream_chat(st.session_state.tokenizer, image_path, input, history,
                                                                max_length=max_length, top_p=top_p,
                                                                temperature=temperature):
        yield response, history


@st.cache_resource
def init_application():
    st.session_state['history'] = []
    st.session_state['chatbot'] = []

    tokenizer = AutoTokenizer.from_pretrained("THUDM/visualglm-6b", trust_remote_code=True)
    model = AutoModel.from_pretrained("THUDM/visualglm-6b", trust_remote_code=True).half().cuda()
    model = model.eval()
    st.session_state.tokenizer = tokenizer
    st.session_state.model = model


def clear_session():
    st.session_state['chatbot'].clear()
    st.session_state['history'].clear()


if "chatbot" not in st.session_state:
    init_application()

st.title('VisualGLM')

uploaded_file = st.file_uploader("Upload Image", type=["png", "jpg", "jpeg"])
container = st.container()

user_input = st.text_input('Input...')
max_length = st.sidebar.slider("Maximum length", 0, 4096, 2048, 1)
top_p = st.sidebar.slider("Top P", 0.0, 1.0, 0.4, 0.01)
temperature = st.sidebar.slider("Temperature", 0.0, 1.0, 0.8, 0.01)

if uploaded_file is not None:
    # image_path = Image.open(uploaded_file)
    image = Image.open(uploaded_file)
    with container:
        st.image(image, use_column_width=True)

    with tempfile.NamedTemporaryFile(delete=False, suffix=".png") as temp:
        image.save(temp.name, "PNG")
        image_path = temp.name
        st.session_state.image = image_path

    if "uploaded_file" in st.session_state and st.session_state.uploaded_file != uploaded_file:
        clear_session()
    st.session_state.uploaded_file = uploaded_file

# 创建两列
col1, col2 = st.columns(2)
send_button = col1.button('🚀   发 送    ')
clear_button = col2.button('🧹 清除历史对话')

if clear_button:
    clear_session()

if send_button:
    if len(user_input) > 0:
        gen = predict(user_input, st.session_state.image, st.session_state['chatbot'], max_length, top_p, temperature,
                      st.session_state['history'] if st.session_state['history'] else [])
        while True:
            try:
                response, st.session_state['history'] = next(gen)
            except StopIteration:  # 当所有的数据都被遍历完，next函数会抛出StopIteration的异常
                st.session_state['chatbot'].append((user_input, response))
                break
        user_input = ''

human_history = st.session_state['chatbot']
for i, (query, response) in enumerate(human_history):
    message_chat(query, avatar_style="big-smile", key=str(i) + "_user")  # User input
    message_chat(response, avatar_style="bottts", key=str(i))  # Model response



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : AI.  @by PyCharm
# @File         : usearch
# @Time         : 2023/8/22 08:49
# @Author       : betterme
# @WeChat       : meutils
# @Software     : PyCharm
# @Description  : 合并【定期】，保存

from meutils.pipe import *
from langchain.vectorstores import USearch as _USearch

from langchain.docstore.document import Document
from langchain.docstore.in_memory import InMemoryDocstore
from langchain.embeddings.base import Embeddings


class USearch(_USearch):

    def similarity_search(
        self,
        query: str,
        k: int = 4,
        threshold: float = 0.5,
        **kwargs: Any,
    ) -> List[Document]:

        docs_and_scores = self.similarity_search_with_score(query, k)

        docs = []
        for doc, score in docs_and_scores:
            score = round(1 - score, 2)
            if score > threshold:
                doc.metadata['score'] = score
                docs.append(doc)
        return docs

    @classmethod
    def from_texts(
        cls,
        texts: List[str],
        embedding: Embeddings,
        metadatas: Optional[List[Dict]] = None,
        ids: Optional[np.ndarray] = None,
        metric: str = "cos",
        **index_kwargs: Any,  # from_documents kwargs
    ) -> _USearch:
        """Construct USearch wrapper from raw documents.
        This is a user friendly interface that:
            1. Embeds documents.
            2. Creates an in memory docstore
            3. Initializes the USearch database
        This is intended to be a quick way to get started.

        Example:
            .. code-block:: python

                from langchain.vectorstores import USearch
                from langchain.embeddings import OpenAIEmbeddings

                embeddings = OpenAIEmbeddings()
                usearch = USearch.from_texts(texts, embeddings)
        """
        embeddings = embedding.embed_documents(texts)

        documents: List[Document] = []
        if ids is None:
            ids = list(map(str, range(len(texts))))
        for i, text in enumerate(texts):
            metadata = metadatas[i] if metadatas else {}
            documents.append(Document(page_content=text, metadata=metadata))

        docstore = InMemoryDocstore(dict(zip(ids, documents)))

        from usearch.index import Index
        index = Index(ndim=len(embeddings[0]), metric=metric, **index_kwargs)
        index.add(np.array(ids), np.array(embeddings))

        return cls(embedding, index, docstore, ids)

    def save_local(self, folder_path: str, index_name: str = "index") -> None:
        """ save index and docstore

        :param folder_path:
        :param index_name:
        :return:
        """

        path = Path(folder_path)
        path.mkdir(exist_ok=True, parents=True)

        self.index.save(path / f"{index_name}.usearch")
        pkl_dump(self.docstore, path / f"{index_name}.pkl")  # 保存 InMemoryDocstore(dict(zip(ids, documents)))

    @classmethod
    def load_local(
        cls,
        folder_path: str,
        embeddings: Embeddings,
        index_name: str = "index",
        **kwargs: Any,
    ) -> _USearch:

        path = Path(folder_path)

        from usearch.index import Index
        index = Index.restore(str(path / f"{index_name}.usearch"))  # os.PathLike对象读取为空
        docstore: InMemoryDocstore = pkl_load(path / f"{index_name}.pkl")

        return cls(embeddings, index, docstore, ids=list(docstore._dict))

    def merge_from(self, target: _USearch) -> None:
        """从0开始建索引"""

        vectors = np.r_[self.index.vectors, target.index.vectors]
        ids = list(map(str, range(len(vectors))))

        self.index.clear()  # 清空重建
        self.index.add(ids, vectors)  # search字符串可能报错

        documents = list(self.docstore._dict.values()) + list(target.docstore._dict.values())

        self.docstore = InMemoryDocstore(dict(zip(ids, documents)))


if __name__ == '__main__':
    from chatllm.llmchain.vectorstores import USearch
    from chatllm.llmchain.embeddings import HuggingFaceEmbeddings

    model_name = '/Users/betterme/PycharmProjects/AI/m3e-small'
    encode_kwargs = {'normalize_embeddings': True, 'show_progress_bar': True, 'device': 'mps'}
    embeddings = HuggingFaceEmbeddings(model_name=model_name, encode_kwargs=encode_kwargs)

    ann = USearch.from_texts(['1', '2', '3'], embeddings, dtype='i8')

    print(ann.similarity_search('1'))



# !/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : AI.  @by PyCharm
# @File         : demo
# @Time         : 2023/12/21 17:22
# @Author       : betterme
# @WeChat       : meutils
# @Software     : PyCharm
# @Description  :

from meutils.pipe import *
from urllib.parse import quote, unquote


_ = quote("sk-api-xyhelper-cn-free-token-for-everyone-xyhelper|| ")

print(_)


print(unquote(_))
